/***************************************************************/
/*   Author: Erwan Loisant                                     */
/*   Date: September 2003                                      */
/*   Licence: General Public Licence                           */
/*                                                             */
/*   Contributors: Todd Rudick, Mathieu Mangeot                */
/***************************************************************/

//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

///////////////////////////////////////////
//   GOBALS
///////////////////////////////////////////

// see dictionaries.js
// var dicos = null;
// var wordDict = null;
// Normally not used (to delete)
// var kanjiDict = null;
// var kanjidic = null;

var strbundle;

var dicts;
var kanjidicts;

var kanjiIndex;
var longestMatch;

var historyBox;
var mojiHistory = Array();
var searched = Object();

var mojiInitialized = false;

/***************************************************************/
/*                Initialization functions                     */
/***************************************************************/

//////////////////////////////////////////////////////////////
// initMoji
//
// PARAMETERS:
//   nothing
// RETURNS:
//   nothing
//
// Load preferences and initialize global variables
//////////////////////////////////////////////////////////////
function initMoji()
{
  var pref = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
  var dictorder = pref.getCharPref("moji.dictorder");
  var kanjidictorder = pref.getCharPref("moji.kanjidictorder");

  try {
    longestMatch = pref.getBoolPref("moji.longestMatch");
  } catch ( e ) {
    longestMatch = false;
  }
  
  kanjiIndex = new Array();
  kanjiIndex["C"] = pref.getBoolPref("moji.kanji.classicalchk");
  kanjiIndex["H"] = pref.getBoolPref("moji.kanji.halpernchk");
  kanjiIndex["N"] = pref.getBoolPref("moji.kanji.nelsonchk");
  kanjiIndex["V"] = pref.getBoolPref("moji.kanji.newnelsonchk");
  kanjiIndex["DR"] = pref.getBoolPref("moji.kanji.fatherjdrchk");
  kanjiIndex["DK"] = pref.getBoolPref("moji.kanji.kanjilearnerchk");
  kanjiIndex["DO"] = pref.getBoolPref("moji.kanji.pgochk");
  kanjiIndex["P"] = pref.getBoolPref("moji.kanji.skipchk");
  kanjiIndex["U"] = pref.getBoolPref("moji.kanji.unihexchk");
  kanjiIndex["I"] = pref.getBoolPref("moji.kanji.tuttledicchk");
  kanjiIndex["IN"] = pref.getBoolPref("moji.kanji.tuttlekanjichk");
  kanjiIndex["Q"] = pref.getBoolPref("moji.kanji.fourchk");
  kanjiIndex["MN"] = pref.getBoolPref("moji.kanji.moro1chk");
  kanjiIndex["MP"] = pref.getBoolPref("moji.kanji.moro2chk");
  kanjiIndex["E"] = pref.getBoolPref("moji.kanji.henshallchk");
  kanjiIndex["K"] = pref.getBoolPref("moji.kanji.gakkenchk");
  kanjiIndex["L"] = pref.getBoolPref("moji.kanji.heisigchk");
  kanjiIndex["O"] = pref.getBoolPref("moji.kanji.pgo2chk");
  kanjiIndex["W"] = pref.getBoolPref("moji.kanji.koreanchk");
  kanjiIndex["Y"] = pref.getBoolPref("moji.kanji.pinyinchk");

  strbundle = window.document.getElementById("strings");
  
  mojiInitialized = true;
  
  var hasDicts = false;
  dicts = new Array();
  kanjidicts = new Array();
  
  if ( dictorder.length > 0 ) {
	var dictsarr = dictorder.split(",");
	  
	for ( var i = 0; i < dictsarr.length; i++ )
	{
	    if ( dictsarr[i].length > 0 ) {
	        try {
			  var edict = new EDict(dictsarr[i]);
		    
			  dicts.push(edict);
			  hasDicts = true;
		    } catch ( e ) {
		      alert(strbundle.getString("dictopenerror") + " " + dictsarr[i] + ": " + e); 
		    }
		}
	}
  }
  
  if ( kanjidictorder.length > 0 ) {
	var dictsarr = kanjidictorder.split(",");
	  
	for ( var i = 0; i < dictsarr.length; i++ )
	{
		if ( dictsarr[i].length > 0 ) {
		  try {
			var kanjidic = new KanjiDict(dictsarr[i]);
		    
			kanjidicts.push(kanjidic);
			hasDicts = true;
		  } catch ( e ) {
		    alert(strbundle.getString("kanjidictopenerror") + " " + dictsarr[i] + ": " + e);
		  }
		}
	}
  }
    
  historyBox = window.document.getElementById('mojihistory');
  
  if ( ! hasDicts ) {
    alert(strbundle.getString("dictRequired"));
    throw strbundle.getString("dictRequired");
  }
}

function closeMoji()
{
  for ( var i = 0; i < dicts.length; i++ )
    dicts[i].close();

  for ( var i = 0; i < kanjidicts.length; i++)
    kanjidicts[i].close();
}

// Get selected text in HTML window
function getSelection() {
   var focusedWindow = document.commandDispatcher.focusedWindow;
   var winWrapper = new XPCNativeWrapper(focusedWindow, 'document', 'getSelection()');
   var selection = winWrapper.getSelection();
   return String(selection);
}

function getWord()
{
  var word = getSelection();
  
  // remove leading and trailing white space.
  word = word.replace(/^\s/g, "");
  word = word.replace(/\s$/g, "");
  
  if (word.length == 0) {
    var sbox = window.document.getElementById('mojitextbox').value;
    sbox = sbox.replace(/\s/g, "");
    
    if ( sbox.length == 0 ) {
      return;
    }
    
    word = sbox;
  }
  
  return word;
}

function lookupKanji()
{
  if ( ! mojiInitialized ) 
    initMoji();

  if ( kanjidicts.length < 1 ) {
    alert(strbundle.getString("dictRequired"));
    return;
  }
  
  var txt = getWord();
  
  if ( txt == null || txt.length == 0 ) {
    alert(strbundle.getString("atleast1kanji"));
    return;
  }

  // var allentries = new Array();
  var firstKanji;
  
  for ( var i = 0; i < txt.length; i++ ) {
    var c = txt.charAt(i);
    
    if ( searched[c] && ( searched[c].kanji != null ) ) {
      if ( firstKanji == null ) {
        firstKanji = historyBox.view.getIndexOfItem(mojiHistory[searched[c].kanji].treeitem);
      }
      continue;
    }

    var histentry = addSearch(c, "kanji");

    if ( histentry != null ) {
      // for ( var j = 0; j < histentry.dict.length; j++ )
      //  for ( var k = 0; k < histentry.dict[j].entries.length; k++ ) 
      //    allentries.push(histentry.dict[j].entries[k].txt);
          
      if ( firstKanji == null ) 
        firstKanji = historyBox.view.getIndexOfItem(histentry.treeitem);      
    }
  }
  
  /* if ( allentries.length > 0 ) {
    var s = "";
  
    for ( var i = 0; i < allentries.length; i++ ) 
    {
      if ( i > 0 )
        s = s + "\r\n";
        
      s = s + allentries[i];
    }
    
    alert("found " + allentries.length + " entries:\r\n" + s);
  } else */
  
  if ( firstKanji == null ) {
    alert(strbundle.getString("atleast1kanji"));
    return;
  } else 
    historyBox.view.selection.select(firstKanji);
}

function clearHist()
{
  var children = window.document.getElementById('histchildren');
  while ( children.hasChildNodes() ) {
    children.removeChild(children.firstChild);
  }
  mojiHistory = Array();
  searched = Object();
}

function exportHist(how)
{
  var text = "";
  
  if ( how == "raw" ) {
    for ( var i = 0; i < mojiHistory.length; i++ ) {
      for ( var j = 0; j < mojiHistory[i].dict.length; j++ ) {
        for ( var k = 0; k < mojiHistory[i].dict[j].entries.length; k++ ) {
          if ( text != "" ) 
            text = text + "\n";
            
          text = text + mojiHistory[i].dict[j].entries[k].txt;
        }
      }
    }

    window.open("data:text/plain;charset=utf-8," + text, "History", "menubar,scrollbars,resizable,dialog=no");
  } else if ( how == "html" ) {
    for ( var i = 0; i < mojiHistory.length; i++ ) {
      if ( text != "" )
        text = text + "\n";
      
      text = text + getHtml(mojiHistory[i], null, null);
    }

    window.open("data:text/html;charset=utf-8," + text, "History");
  }  
}

function historySelect()
{
  var s = historyBox.view.selection.currentIndex;

  if ( s < 0 )
    return;
  
  historyBox.treeBoxObject.ensureRowIsVisible(s);
  
  // alert("selected treeitem (" + s + "): " + historyBox.view.getItemAtIndex(s).id);

  var id = historyBox.view.getItemAtIndex(s).id;
  
  var histid = id.match(/^hist([0-9]+)(\.([0-9]+))?(\.([0-9]+))?/);
  
  if ( histid[1] == null ) {
    throw "regexp does not return histentno";
    return;
  }
  
  var histno = histid[1];
  var dictno = histid[3];
  var matchno = histid[5];
  
  if ( histno == null || (histno < 0) || (histno >= mojiHistory.length) ) {
    throw "Histno " + histno + " out of bounds (0:" + mojiHistory.length + ")";
  }
  
  var html = getHtml(mojiHistory[histno], dictno, matchno);  
  window.document.getElementById('kanjiframe').setAttribute('src', "data:text/html;charset=utf-8," + html);
}

// lookupWord()
//
// search for a word in the word dictionaries.
//
function lookupWord()
{
  if ( ! mojiInitialized ) 
    initMoji();
  
  if ( dicts.length < 1 ) {
    alert(strbundle.getString("dictRequired"));
    return;
  }
  
  // O.K. figure out which word we have to look for.
  var txt = getWord();

  // nothing to do...
  if ( txt == null || txt.length == 0 ) {
    alert(strbundle.getString("selectword"));
    return;
  }
  
  var histselitem;
  var rest = "";
  
  while ( txt.length > 0 ) {
    // check if we have search for this term before. If yes, just select the corresponding top entry
    // in the history box which will show the entry in turn.
    if ( searched[txt] && ( searched[txt].word != null ) ) {
      if ( histselitem == null ) 
        histselitem = mojiHistory[searched[txt].word];
      
      txt = rest;
      rest = "";
      continue;
    }

    var histentry = addSearch(txt, "word");
    
    if ( histentry != null && histselitem == null ) {
      histselitem = histentry;
    }
    
    if ( longestMatch ) {
      if ( histentry == null ) {
        if ( txt.length > 1 ) {
          rest = txt.substr(txt.length - 1) + rest;
          txt = txt.substr(0, txt.length - 1);
        } else {
          txt = rest;
          rest = "";
        }
      } else {
        txt = rest;
        rest = "";
      }
    } else {
      txt = "";
    }
    
    // DEBUG debugging code to dump all the matches in the history entry.
    /* if ( histentry != null ) {
      var s = "";
      var c = 0;
    
      for ( var i = 0; i < histentry.dict.length; i++ ) 
      {
        for ( var j = 0; j < histentry.dict[i].entries.length; j++ )
        {
          if ( c > 0 )
            s = s + "\r\n";
            
          c = c + 1;
          
          s = s + histentry.dict[i].entries[j].txt;
          
          if ( histentry.dict[i].entries[j].reason ) 
            s = s + " [" + histentry.dict[i].entries[j].reason + "]";
        }
      }
      
      alert("found " + c + " entries in " + histentry.dict.length + " dictionaries:\r\n" + s);
    } else
      alert("no entries found");
    */ 
  }
  
  if ( histselitem == null ) {
        var sorry = "<html>" +
                '<head><link rel="stylesheet" type="text/css" href="chrome://moji/skin/output.css" /></head>' +
                "<body><div style='text-align: center'><p>"+ strbundle.getString("sorry") +"<br/>" +
	        "<span class='lookedword'>" + txt + "</span><br/>" +
		strbundle.getString("notfound")+"</p><br/>" +
		"<span class='kaomoji'>m(_ _)m</span>" +
		"</body></html>";
    window.document.getElementById('kanjiframe').setAttribute('src', "data:text/html;charset=utf-8," + sorry);
  } else {
    // and select it which will have the entry rendered in the sidebox.
    historyBox.view.selection.select(historyBox.view.getIndexOfItem(histselitem.treeitem));
  }
}

// addSearch does a search for txt in the dict class cls.
// cls can be "word" or "kanji" at this time.
// addSearch returns the history entry added to mojiHistory or null if nothing has been found.
function addSearch(txt, cls)
{
  // O.K. txt contains the text we are searching for. Prepare the new history entry 
  // which goes into mojiHistory in the end.  
  var histentry = new Object();
  var searchdicts;
  
  // remember the searchword and which class we are in.
  histentry.searchword = txt;
  histentry.cls = cls;

  if ( cls == "word" ) 
    searchdicts = dicts;
  else if ( cls == "kanji" ) 
    searchdicts = kanjidicts;
  else
    throw "Unknown class \"" + cls + "\"";
      
  // and prepare the array for the dictionaries.
  histentry.dict = new Array();
  
  // now we run through all active dictionaries and search for the word in each of them.
  for ( var i = 0; i < searchdicts.length; i++ ) {
    // get all matches and derived/conjugated matches (if conjugation file is present)
    // from the dictionary.
    
    var e = searchdicts[i].search(txt);
    if ( e != null && e.length > 0 ) {
      // O.K. we have found something in this dictionary. prepare the dictionary entry.
      var histentdict = new Object();
      // save the name of the dictionary of later.
      histentdict.name = searchdicts[i].name;
      
      // now we have to process each of the found entries, generate the html code
      for ( var j = 0; j < e.length; j++ ) {
        var fm = getFirstMeaning(e[j].txt, cls);

        // this is the first match in this dictionary thus the first meaning of this match
        // is also considered as first meaning for the whole dictionary.
        if ( j == 0 ) {
          histentdict.firstMeaning = fm;
        }
        
        // create the html and remember the first meaning for this match.
        if ( cls == "word" ) {
          e[j].html = formatDictEntry(e[j]);
        } else if ( cls == "kanji" ) {
          e[j].html = formatKanjiDictEntry(searchdicts[i], e[j]);
        } else {
          throw "Unknown class \"" + cls + "\"";
        }
        e[j].firstMeaning = fm;
      }
      
      // O.K. almost done. We need a first meaning for the whole search as well, which 
      // we establish here if this is actually the first dictionary we found matches in.
      if ( histentry.dict.length == 0 ) {
        histentry.firstMeaning = histentdict.firstMeaning;
      }

      // save the new object and push it onto the dict array in the history entry.
      histentdict.entries = e;
      histentry.dict.push(histentdict);
    }
  }
  
  // if we found no matches the dicts array in the history entry is empty.
  if ( histentry.dict.length == 0 )
  {
    return null;
  } else {
    // Found matches. Store the new entry in the mojiHistory.
    var ind = mojiHistory.length;
    mojiHistory.push(histentry);
    
    // also remember that we have searched for this word.
    if ( ! searched[txt] ) 
      searched[txt] = new Object();
    
    if ( cls == "word" ) 
      searched[txt].word = ind;
    else if ( cls == "kanji" )
      searched[txt].kanji = ind;
    else
      throw "Unknown class \"" + cls + "\"";
    
    // add the entry to the tree.
    addHistoryEntry(histentry, ind);
    
    return histentry;
  }
}

function getFirstMeaning(txt, cls)
{
  if ( cls == "word" ) {
    // This regexp extracts the first meaning entry from the dictionary entry.
    // the first meaning is used in the tree view next to the search word.
    // It assumes the following structure of an entry:
    // "word [reading] explanations"
    // explanation is an arbitrary list of entries each of them between slashes "/", e.g.
    //   /a/b/c/d/e/f/
    // The first meaning is the first meaningful entry. There are a couple of structural
    // and grammatical note entries which we discard. Those are:
    // -  e.g. "[1]", which numbers various different meanings for one word.
    // -  e.g. "{Zool.}", which explains the source or origin of the word or the area of use.
    // -  e.g. "(schriftspr.)", when it is used e.g. in written language only.
    // 
    // thus the regexp consists of the following parts:
    // "^[^ ]* " matching the beginning of the entry up to the first space = word
    // "(\[[^\]]*\] )?" an optional reading entry between [] followed by a space.
    // "(\/(\{[^}]*\}|\[[^\]*]\]|\([^\)]*\)))*" an arbitrary+optional list of note entries [], {} and ()
    //                each starting with a slash
    // "\/([^\/]*)/)" that is the first entry then that does not match the one above = first meaning
    var fm = txt.match(/^[^ ]* (\[[^\]]*\] )?(\/(\{[^}]*\}|\[[^\]*]\]|\([^\)]*\)))*\/([^\/]*)/)[4];
    
    // the english dictionary for instance does contain additional grammatical information in
    // the first explanation, e.g. (n) for nouns. These are removed here.
    fm = fm.replace(/\([a-zA-Z0-9\-]+\)/g,"");
    return fm;
  } else if ( cls == "kanji" ) {
    return txt.match(/{([^}]*)}/)[1];
  } else
    throw "Unknown class \"" + cls + "\"";
}

// getHtml returns the html code for a selected history entry.
// histent is the actual history entry which has been selected.
// dictno is the dictionary index in case a sub entry has been selected in the tree.
// matchno is the match in the dictionary dictno that has been selected.
function getHtml(histent, dictno, matchno)
{
  var html = '<html>\n<head><link rel="stylesheet" type="text/css" href="chrome://moji/skin/output.css" /></head>\n<body>\n';
  // the iframe only compares some xxx first bytes of the src tag to decide whether or not
  // the contents has changed and needs to be repainted. Therefore switching between a tree node
  // and its direct parent may result in no change as the first entry is always the identical
  // parent node. For this reason we just a random number here in the beginning to make sure
  // the (most likely ;-) each entry is unique.
  html += "<!-- " + Math.random() + " -->";
  var dict1, dict2;
  
  if ( dictno == null ) {
    dict1 = 0;
    dict2 = histent.dict.length - 1;
  } else {
    dict1 = dictno;
    dict2 = dictno;
  }

  for ( var i = dict1; i <= dict2; i++ ) {
    var match1 = 0;
    var match2 = histent.dict[i].entries.length - 1;
    
    if ( matchno != null ) { 
      if ( matchno >= 0 && matchno <= match2 ) {
        match1 = matchno;
        match2 = matchno;
      } else
        continue; // skip this one.
    }

    // all we have to do is copy all the formatted entries in .html together.    
    for ( var j = match1; j <= match2; j++ ) {
      html += histent.dict[i].entries[j].html;
    }
  }
    
  html += "</body></html>";
  
  return html;
}

// add the necessary items to the history tree.
// - histent is the moji history entry to be added.
// - ind defined the new index of the entry to be used.
//
// This function is a little bit tricky as the structure of the tree changes
// depending on the number of matches found. The basic structure is:
// + word dict1.match1
// +-+ word dict1.match1
// +-+-+ word dict1.match1
// +-+-+ word dict1.match2
// +-+ word dict2.match1
// +-+-+ word dict2.match1
// +-+-+ word dict2.match2
// i.e. a higher level entry always duplicates the entry information of the first child therefore
//   if you have the subtree hidden you always see the first match meaning on the subtree trunk.
//
// However, for optimization and reduction of unnecessary subtrees, no subtree will be added if
// it would contain only one entry. That means if there is only one match in a dictionary the 
// dictionary entry have no subtree and it represents itself the only match found. Likewise,
// If matches are only found in one dictionary there won't be the intermediate dict1, dict2, ...
// structure but instead the dictionary matches will be added as subtree to the trunk.
// Finally, if there is actually only one match in one dictionary the trunk itself will 
// represent this match and there won't be a subtree to this search word.
//
// Each match is basically represented in a <treerow>
// If a subtree is required this is represented by a <treechilden> node which
//  in turn must contain a <treeitem> for each child. (Which in turn would
//  contain the <treerow> if it is a leave node).
function addHistoryEntry(histent, ind)
{
  var children = window.document.getElementById('histchildren');
  
  // create the top tree item first. This represents the whole search.
  var topitem = window.document.createElement("treeitem");
  histent.treeitem = topitem;
  topitem.setAttribute("id", "hist" + ind);
  topitem.setAttribute("open", "false");
  children.appendChild(topitem);
  
  // if there is more than one dictionary with matches we have to create a subtree
  // for the dictionaries.
  var top1, main1;
  if ( histent.dict.length > 1 ) {
    top1 = window.document.createElement("treechildren");
  }
    
  for ( var i = 0; i < histent.dict.length; i++ ) {
    // if there is more than one match in this dictionary we have to create
    // a subtree for all matches.
    var multi = (histent.dict[i].entries.length > 1);
    var top2, main2;
    
    if ( multi ) {
      top2 = window.document.createElement("treechildren");
    }
    
    for ( var j = 0; j < histent.dict[i].entries.length; j++ ) {
      // create the treerow element for this match.
      var row = window.document.createElement("treerow");

      var cell = window.document.createElement("treecell");
      cell.setAttribute("ref", "histsearchword");
      cell.setAttribute("label", histent.searchword);
      cell.setAttribute("properties", histent.cls);
      row.appendChild(cell);
      
      var cell = window.document.createElement("treecell");
      cell.setAttribute("ref", "histmeaning");
      cell.setAttribute("label", histent.dict[i].entries[j].firstMeaning);
      cell.setAttribute("properties", histent.cls);
      row.appendChild(cell);
      
      var cell = window.document.createElement("treecell");
      cell.setAttribute("ref", "histdict");
      cell.setAttribute("label", histent.dict[i].name);
      cell.setAttribute("properties", histent.cls);
      row.appendChild(cell);
      
      if ( histent.dict[i].entries[j].reason ) {
        var cell = window.document.createElement("treecell");
        cell.setAttribute("ref", "histreason");
        cell.setAttribute("label", histent.dict[i].entries[j].reason);
        cell.setAttribute("properties", histent.cls);
        row.appendChild(cell);
      }

      // if there are several matches for this dictionary, we have to wrap
      // this row in a treeitem element which we will add to the treechildren
      // element which has been created above in top2.
      if ( multi ) {
        var item = window.document.createElement("treeitem");
        item.setAttribute("id", "hist" + ind + "." + i + "." + j);
        item.appendChild(row);
        top2.appendChild(item);
        
        // the root of this subtree duplicates the first match.
        if ( j == 0 )
          main2 = row.cloneNode(true);
      } else {
        // it is only one match thus there will be only one row which will be passed on
        // after this loop.      
        top2 = row;
      }
    }
    
    // top2 now represents the node we have to put into the tree for all matches in this dictionary.
    // top2 can be either a <treerow> element if there is only one match or a <treechilden> element
    // if there are several matches.
    // The follow ifs branch in all four possible cases we have to consider:
    // - several dicts, several matches
    // - several dicts, one match
    // - one dict, several matches
    // - one dict, one match
    // To keep the code clearer I have kept those cases seperate and did not optimize it...
    if ( multi ) {
      if ( histent.dict.length > 1 ) {
        // we have several dicts and this dict has several matches. These matches go
        // into a subtree...
        var item = window.document.createElement("treeitem");
        item.setAttribute("id", "hist" + ind + "." + i);
        item.setAttribute("container", "true");
        item.setAttribute("open", "false");
        
        // first we add the cloned entry which represents the whole subtree and then
        // we add the whole subtree structure.
        item.appendChild(main2);
        item.appendChild(top2);
        
        // and put it in there.
        top1.appendChild(item);
      } else {
        // There is only one dictionary but we have several matches.
        topitem.setAttribute("container", "true");
        topitem.appendChild(main2);
        topitem.appendChild(top2);
      }
      
      // we have to clone a the first meaning node for the whole tree.
      if ( i == 0 )
        main1 = main2.cloneNode(true);
    } else {
      // O.K. there is only one match here...
      
      if ( histent.dict.length > 1 ) {
        // but several dicts for it, so we only have to add the treerow element here...
        var item = window.document.createElement("treeitem");
        item.setAttribute("id", "hist" + ind + "." + i);
        item.setAttribute("open", "false");
        item.appendChild(top2);
        top1.appendChild(item);
        
        if ( i == 0 )
          main1 = top2.cloneNode(true);
      } else {
        // one dict, one match, then this is it...
        topitem.appendChild(top2);
      }
    }
  }
  
  // if we have several dicts we have to fill the topitem here...
  if ( histent.dict.length > 1 ) {
    topitem.appendChild(main1);
    topitem.appendChild(top1);
    topitem.setAttribute("container", "true");
  }
}

// Replace lexical abbreviations by full words. TODO: go multilingual.
function lexicalTerm(input, abbreviation, title) {
  return input.replace(new RegExp("\\\(" + abbreviation + "\\\)", "g"),
  		    '<span class="abbrev"><a href=\'javascript:newWin=window.open("","'+abbreviation+'","menubar=no,width=300,height=150");newWin.document.write("<center>'+title+'</center>");\'>'+abbreviation+'</a>. </span>');
}

// Generate HTML output for the [part, i] entry
function formatDictEntry(entry) {
  var tstring = entry.txt.replace(
              /^([^ ]*) (\[([^\]]*)\] )?\/(.*)\//,"$1\n$3\n$4"
              ).split("\n");

  var ret="";

  txt = tstring[0];

  var meanings = tstring[2].replace(/\//g, "; ");
  
  // Interpret lexical terms     LOCALE 
  meanings = this.lexicalTerm(meanings, 'abbr', 'abbreviation');
  meanings = this.lexicalTerm(meanings, 'adj', 'adjective');
  meanings = this.lexicalTerm(meanings, 'adv', 'adverb');
  meanings = this.lexicalTerm(meanings, 'adv-n', 'adverbial noun');
  meanings = this.lexicalTerm(meanings, 'adj-na', 'adjectival noun or quasi-adjectives');
  meanings = this.lexicalTerm(meanings, 'adj-no', 'noun which may take the genetive case particule "no"');
  meanings = this.lexicalTerm(meanings, 'adj-pn', 'pre-noun adjective');
  meanings = this.lexicalTerm(meanings, 'adj-s', 'special adjective');
  meanings = this.lexicalTerm(meanings, 'arch', 'archaism');
  meanings = this.lexicalTerm(meanings, 'ateji', 'ateji reading of the kanji');
  meanings = this.lexicalTerm(meanings, 'aux', 'auxiliary word or phrase');
  meanings = this.lexicalTerm(meanings, 'aux-v', 'auxiliary verb');
  meanings = this.lexicalTerm(meanings, 'conj', 'conjuction');
  meanings = this.lexicalTerm(meanings, 'col', 'colloquialism');
  meanings = this.lexicalTerm(meanings, 'exp', 'expression');
  meanings = this.lexicalTerm(meanings, 'ek', 'exclusively kanji, rarely just in kana');
  meanings = this.lexicalTerm(meanings, 'fam', 'familiar language');
  meanings = this.lexicalTerm(meanings, 'fem', 'female term or language');
  meanings = this.lexicalTerm(meanings, 'gikun', 'gikun (meaning) reading');
  meanings = this.lexicalTerm(meanings, 'gram', 'grammatical term');
  meanings = this.lexicalTerm(meanings, 'hon', 'honorific or respectful (sonkeigo) language');
  meanings = this.lexicalTerm(meanings, 'hum', 'humble (kenjougo) language');
  meanings = this.lexicalTerm(meanings, 'id', 'idiomatic expression');
  meanings = this.lexicalTerm(meanings, 'int', 'interjection');
  meanings = this.lexicalTerm(meanings, 'iK', 'word containing irregular kanji usage');
  meanings = this.lexicalTerm(meanings, 'ik', 'word containing irregular kana usage');
  meanings = this.lexicalTerm(meanings, 'io', 'irregular okurigana usage');
  meanings = this.lexicalTerm(meanings, 'MA', 'martial arts term');
  meanings = this.lexicalTerm(meanings, 'male', 'male term or language');
  meanings = this.lexicalTerm(meanings, 'm-sl', 'manga slang');
  meanings = this.lexicalTerm(meanings, 'n', 'common noun');
  meanings = this.lexicalTerm(meanings, 'n-adv', 'adverbial noun');
  meanings = this.lexicalTerm(meanings, 'n-t', 'noun (temporal)');
  meanings = this.lexicalTerm(meanings, 'n-suf', 'noun, used as suffix');
  meanings = this.lexicalTerm(meanings, 'n-pref', 'noun, used as prefix');
  meanings = this.lexicalTerm(meanings, 'neg', 'negative (in a negative sentence, or with negative verb)');
  meanings = this.lexicalTerm(meanings, 'neg-v', 'negative verb (when used with)');
  meanings = this.lexicalTerm(meanings, 'num', 'number, numeric');
  meanings = this.lexicalTerm(meanings, 'obs', 'obsolete term');
  meanings = this.lexicalTerm(meanings, 'obsc', 'obscure term');
  meanings = this.lexicalTerm(meanings, 'oK', 'word containing out-dated kanji');
  meanings = this.lexicalTerm(meanings, 'ok', 'out-dated or obsolete kana usage');
  meanings = this.lexicalTerm(meanings, 'pol', 'polite (teineigo) language');
  meanings = this.lexicalTerm(meanings, 'pref', 'prefix');
  meanings = this.lexicalTerm(meanings, 'prt', 'particle');
  meanings = this.lexicalTerm(meanings, 'qt', 'quod vide');
  meanings = this.lexicalTerm(meanings, 'sl', 'slang');
  meanings = this.lexicalTerm(meanings, 'suf', 'suffix');
  meanings = this.lexicalTerm(meanings, 'uK', 'word usually written using kanji alone');
  meanings = this.lexicalTerm(meanings, 'uk', 'word usually written using kana alone');
  meanings = this.lexicalTerm(meanings, 'v1', 'Ichidan verb');
  meanings = this.lexicalTerm(meanings, 'v5', 'Godan verb');
  meanings = this.lexicalTerm(meanings, 'v5u', 'Godan verb u');
  meanings = this.lexicalTerm(meanings, 'v5u-s', 'Godan verb u (special class)');
  meanings = this.lexicalTerm(meanings, 'v5k', 'Godan verb ku');
  meanings = this.lexicalTerm(meanings, 'v5g', 'Godan verb gu');
  meanings = this.lexicalTerm(meanings, 'v5s', 'Godan verb su');
  meanings = this.lexicalTerm(meanings, 'v5t', 'Godan verb tsu');
  meanings = this.lexicalTerm(meanings, 'v5n', 'Godan verb nu');
  meanings = this.lexicalTerm(meanings, 'v5b', 'Godan verb bu');
  meanings = this.lexicalTerm(meanings, 'v5n', 'Godan verb mu');
  meanings = this.lexicalTerm(meanings, 'v5r', 'Godan verb ru');
  meanings = this.lexicalTerm(meanings, 'v5k-s', 'Godan verb - Iku/Yuku special class');
  meanings = this.lexicalTerm(meanings, 'v5z', 'Godan verb - -zuru special class (alternative form of -jiru verbs)');
  meanings = this.lexicalTerm(meanings, 'v5aru', 'Godan verb - -aru special class');
  meanings = this.lexicalTerm(meanings, 'v5uru', 'Godan verb - Uru old class verb (old form of Eru)');
  meanings = this.lexicalTerm(meanings, 'vi', 'intransitive verb');
  meanings = this.lexicalTerm(meanings, 'vs', 'noun or participle which takes the aux. verb suru');
  meanings = this.lexicalTerm(meanings, 'vs-i', 'suru verb - irregular');
  meanings = this.lexicalTerm(meanings, 'vs-s', 'suru verb - special class');
  meanings = this.lexicalTerm(meanings, 'vk', 'Kuru verb - special class');
  meanings = this.lexicalTerm(meanings, 'vt', 'transitive verb');
  meanings = this.lexicalTerm(meanings, 'vulg', 'vulgar expression or word');
  meanings = this.lexicalTerm(meanings, 'X', 'rude or X-rated term');
      
  ret += '<p>';
  
  if ( entry.reason ) {
    ret += '<span class="reason">' + strbundle.getString("inflexion") + ' (' + entry.reason + ')</span><br/>'; // LOCALE of reason
  }

  ret += '<span class="wordinkanji">' + tstring[0] + '</span>';
  
  if ( tstring[1] != "" ) {
    ret += '&nbsp;<span class="yomi">[' + tstring[1] + ']</span>'
  }

  ret += '<br/><span class="translation">&nbsp;&nbsp;&nbsp;'+ meanings + "</span></p>";

  return ret;
};

function formatKanjiDictEntry(kanjidict, entry)
{
  var txt = entry.txt;
  
  var html = "<p align='center'><table width='95%'><tr>";
  html += "<td align='center' valign='top'>";
  
  html += "<span class=\"kanji\">" + entry.key + "</span>";
  
  html += "<div class=\"eigo\"><ul>";
  for ( var i = 0; i < entry.meanings.length; i++ )
    html += "<li>" + entry.meanings[i] + "</li>";
  html += "</ul></div>";
  
  html += "</td>";

  var bushu="";
  if ( entry.info["B"] ) {
    bushu = kanjidict.radicals.radicalData[entry.info["B"]] + " " + entry.info["B"];
  }
  
  var grade = strbundle.getString("uncommon");
  if ( entry.info["G"] ) {
    grade = entry.info["G"];
    var numnames = strbundle.getString("numlist").split(/ /);
    
    if ( ( grade > 0 ) && ( grade < 8 ) ) grade = numnames[grade-1];
    if ( grade == "8" ) grade = strbundle.getString("genuse");
    if ( grade == "9" ) grade = strbundle.getString("names");
  }
  
  var freq = "-";
  if ( entry.info["F"] ) {
    freq=entry.info["F"];
  }
  
  var strokes = "-";
  if ( entry.info["S"] ) {
    strokes = entry.info["S"];
  }
  
  html += "<td valign='top' align='center'>";
  html += "<table border='1' cellpadding='1' cellspacing='0'><tr>";
  html += "<td class=\"infotxt\"><b>"+strbundle.getString("radical")+"</b><br>"+bushu+"</td>";
  html += "<td class=\"infotxt\"><b>"+strbundle.getString("grade")+"</b><br>"+grade+"</td>";
  html += "</tr><tr>";
  html += "<td class=\"infotxt\"><b>"+strbundle.getString("freq")+"</b><br>"+freq+"</td>";
  html += "<td class=\"infotxt\"><b>"+strbundle.getString("strokes")+"</b><br>"+strokes+"</td>";
  html += "</tr></table>";  
  html += "</td>";
  
  html += "</tr>";
  
  if ( kanjidict.radicals.radkfileData[entry.key] ) {
    var rads = "";

    for ( var prop in kanjidict.radicals.radkfileData[entry.key] ) {
      if ( prop != entry.key ) {
        rads += "<tr><td><span style='color:#3333FF'>" + prop + "</span>";
        
        if ( ( kanjidict.radicals.radicalReading[prop] )
             && ( kanjidict.radicals.radicalMeaning[prop] ) ) {
          rads += " </td><td class=\"radicalreading\">";
          rads += kanjidict.radicals.radicalReading[prop];
          rads += "</td><td class=\"radicalmeaning\">";
          rads += kanjidict.radicals.radicalMeaning[prop];
        } else {
          var radentry = kanjidict.search(prop);

          if ( radentry != null && radentry.length > 0 ) {
            var defs = radentry[0].txt.split(/[{}]/);
            if ( defs.length >= 2 ) {
                rads += " </td><td colspan=2 class=\"radicalmeaning\">" + defs[1];
            }
          }
        }
        rads += "</td></tr>";
      }
    }
    
    if ( rads != "" ) {
      html += "<tr width='100%'><td colspan='2' width='100%'><table class=\"rads\">"
      html += rads + "</table></td></tr>";
    }
  }

  var readings = "";
  
  for ( var i = 0; i < entry.reading.length; i++ ) {
    var r = entry.reading[i].replace(/\.(.*)/, "<span style='color:888888'>$1</span>");
    
    if ( i == 0 )   
      readings += r;
    else
      readings += '&#12539;' + r;
  }
  
  for ( var i = 0; i < entry.nameReading.length; i++ ) {
    var r = entry.nameReading[i].replace(/\.(.*)/, "<span style='color:888888'>$1</span>");
  
    if ( i == 0 ) {
      readings += "<br><span class=\"reading\"><u>" + strbundle.getString("namereadings") + "</u></span><br>"
      readings += r;
    } else
      readings += '&#12539;' + r;
  }
  
  for ( var i = 0; i < entry.radicalReading.length; i++ ) {
    var r = entry.radicalReading[i].replace(/\.(.*)/, "<span style='color:888888'>$1</span>");

    if ( i == 0 ) {
      readings += "<br><span class=\"reading\"><u>" + strbundle.getString("radicalreadings") + "</u></span><br>"
      readings += r;
    } else
      readings += '&#12539;' + r;
  }
  
  html += "<tr><td colspan='2'>";
  html += "<div class=\"kdata\">" + readings + "</div>"

  var details = "";
  
  // TODO localize or put in chrome file
  var DetailControl = {
    "C": "Classical Radical",
    "H": "Halpern Index",
    "N": "Nelson Index",
    "V": "New Nelson Index",
    "DR": "Father Joseph De Roo Index",
    "DK": "Kanji Learners Dictionary Index",
    "DO": "P.G. O'Neill Index",
    "P": "Skip Pattern",
    "U": "Unicode Hex Value",
    "I": "Tuttle Kanji Dictionary",
    "IN": "Tuttle Kanji & Kana",
    "Q": "Four Corner Code",
    "MN": "Morohashi Daikanwajiten Index",
    "MP": "Morohashi Daikanwajiten Volume/Page",
    "E": "Henshall Index",
    "K": "Gakken Kanji Dictionary Index",
    "L": "Heisig Index",
    "O": "P.G. O'Neill Japanese Names Index",
    "W": strbundle.getString("koreanreading"),
    "Y": "PinYin" };
    
  var colors=["#FFFFFF", "#DDDDDD"];
  var i = 0;
  
  for (var prop in DetailControl ) {
    if ( entry.info[prop] && kanjiIndex[prop] ) {
      details += "<tr style='background-color:";
      details += colors[i % colors.length] + ";'>";
      details += "<td class=\"detail\">" + DetailControl[prop] + "</td>";
      details += "<td class=\"detail\">" + entry.info[prop] +"</td></tr>";
      i++;
    }
  }

  if (details!="") details="<table>"+details+"</table>";

  html += details + "</td></tr>";
  html += "</table>";
  
  return html;
}